#Computes the likelihood ratio statistics from Davis, Huang and Yao (1995), "Testing for
#a change in the parameter values and order of an autoregressive process", Annals of
#Statistics 23, 282-304.
LRT<-function(tdat, p, n = length(tdat)){
#tdat = data vector of length n
#p = order of AR model
#sig2hat = variance estimator
#Q = storage matrix for Q(p+1, tau) and Q(tau + 1, n)
#lrtStat = matrix of time points and associated raw LRT statistics
  sig2hat=0
  Q <- matrix(0, n-2*p-1, 2)
#initialize XTX^{-1}, XTx and xTx for Q(p+1, tau)
  XTXInvLow <- initXTX(tdat,2*p,p)
  XTxLow <- t(XTXInvLow)%*%as.matrix(tdat[(p+1):(2*p+1)])
  XTXInvLow <- solve(t(XTXInvLow)%*%XTXInvLow)
  xTxLow <- t(as.matrix(tdat[(p+1):(2*p+1)]))%*%as.matrix(tdat[(p+1):(2*p+1)])
#initialize XTX^{-1}, XTx and xTx for Q(tau+1, n)
  XTXInvUp <- initXTX(tdat,n-1,p)
  XTxUp <- t(XTXInvUp)%*%as.matrix(tdat[(n-p):n])
  XTXInvUp <- solve(t(XTXInvUp)%*%XTXInvUp)
  xTxUp <- t(as.matrix(tdat[(n-p):n]))%*%as.matrix(tdat[(n-p):n])
  
  for (t in (2*(p+1)):n){
#update XTX^{-1}, XTx and xTx for Q(p+1, t)
    xLow <- rbind(1,as.matrix(tdat[(t-1):(t-p)],p,1))
    XTXInvLow <- updateXTXinv(XTXInvLow,xLow)
    XTxLow <- updateXTx(XTxLow,tdat,t,p)
    xTxLow <- xTxLow + tdat[t]^2
#repeat the process for Q(t+1, n)
    xUp <- rbind(1,as.matrix(tdat[(n-t+p):(n-t+1)],p,1))
    XTXInvUp <- updateXTXinv(XTXInvUp,xUp)
    XTxUp <- updateXTx(XTxUp,tdat,n-t+p+1,p)
    xTxUp <- xTxUp + tdat[n-t+p+1]^2
#QLow = Q(p+1, t) and QUp = Q(t+1, n)
    QLow <- xTxLow - t(XTxLow)%*%XTXInvLow%*%XTxLow
    QUp <- xTxUp - t(XTxUp)%*%XTXInvUp%*%XTxUp
    Q[t-2*p-1, 1:2] <- c(QLow, QUp)
  }
#compute the variance estimator
  sig2hat <- Q[n-2*p-1,1]/(n-p)
#now combine the element of Q to obtain the raw LRT statistics
  lrtStat <- matrix(0,n-3*p-3,2)
  for(t in (2*p+2):(n-p-2)){
    lrtStat[t - (2*p+1),1] <- t
    lrtStat[t - (2*p+1),2] <- (Q[n-(2*p+1),1] - Q[t-(2*p+1),1] - Q[n-t-(p+1),2])
  }
#the LRT statistic is the largest of the elements in tStat[,2]
  tStat <- max(lrtStat[,2])
  outList <- list("test statistic" = tStat,
                  "change point estimator" = lrtStat[which(lrtStat[,2] == tStat, T),1],
                  "sig2hat" = sig2hat,
                  "raw test statistics" = lrtStat)
  outList
}

updateXTXinv<-function(A, x){
  temp <- A%*%x
  d <- as.numeric(1 + t(x)%*%A%*%x)
  return(A - temp%*%t(temp)/d)
}

updateXTx<-function(A, x, t, p){
  temp <- matrix(c(1,x[(t-1):(t-p)]),p+1,1)
  A <- A + temp*as.numeric(x[t])
}

initXTX<-function(x, t, p){
#initializes the XTX matrix for recursion from above and below
#t is the upper index 
  A <- matrix(0, p+1, p+1)
#first column is the unit vector
  A[,1] <- 1
  for (i in 1:(p+1)){
    A[i,2:(p+1)] <- x[(t-p+i-1):(t-2*p+i)]
  }            
  A
}

####################################################################################
#function to generate the data used in the examples

datGen <- function(n, d1 = 0, d2 = 0, d3 = 0){
  x<-matrix(0,n,1)
  x[1]<-.752 + rnorm(1)
  x[2]<-.753 + .2203*rnorm(1)
  for (i in 3:(floor(n/2))) {
    x[i]<- .753 + (0.2203)*x[i - 1] +(0.0751)*x[i - 2] + rnorm(1)
  }
  for (i in (floor(n/2)+1):n) {
    x[i]<- (.753 + d1) + (0.2203 + d2)*x[i - 1] +
      (0.0751 + d3)*x[i - 2] + rnorm(1)                        
  }
  x
}

####################################################################################
#example data sets were produced with the following code

#null case
set.seed(123)
x <- datGen(1000)

#alternative
set.seed(123)
x <- datGen(100, 1, .05, .07)
